import std.socket;
import std.socketstream;

import std.stream;
import std.string;
import std.ctype;
import std.uri;
import std.conv;
import std.c.stdio;

class HttpError : Error {
    this(char[] msg) { super(msg); }
}

class HttpResponse {
public:
    this(char[] data) {
        all = messageBody = data;
    }

    this(Stream data) {
        statusLine = data.readLine();

        int index = 0;
        while (index < statusLine.length &&
               !isspace(statusLine[index]))
            index++;

        // it's HTTP/0.9
        if (index < 4 || statusLine[0..4] != "HTTP") {
            messageBody = statusLine ~ "\n";
            statusLine = null;
            try {
                char c;
                while ((c = data.getc()) != 255) {
                    messageBody ~= c;
                }
            }
            // hack!
            catch (Object re) {}
//            catch (ReadError re) {}

            return;
        }

        httpVersion = statusLine[0..index];

        while (isspace(statusLine[index])) index++;
        int index2 = index;

        while (!isspace(statusLine[index])) index++;
        code = statusLine[index2..index];

        while (isspace(statusLine[index])) index++;
        message = statusLine[index..statusLine.length];

        int len = -1;
        bool transfer_encoding_chunked = false;
        char[] line = data.readLine();
        while (line != "") {
            int f = find(line, ": ");
            char[] k = line[0..f];
            char[] v = line[f+2..line.length];

            char[] lk = std.string.tolower(k);
            if (lk == `content-length`) {
                len = toInt(v);
            }
            else if (lk == `transfer-encoding`) {
                if (std.string.tolower(v) == `chunked`) {
                    transfer_encoding_chunked = true;
                }
            }

            header[k] = v;

            line = data.readLine();
        }

        if (transfer_encoding_chunked) {
            messageBody = "";
            char[] lenStr = data.readLine();
            sscanf(toStringz(lenStr), "%x", &len);
            while (len) {
                messageBody ~= data.readString(len);
                lenStr = data.readLine();
                lenStr = data.readLine();
                sscanf(toStringz(lenStr), "%x", &len);
            }
            lenStr = data.readLine();
        }
        else if (len == -1) {
            static const int BUFSIZE = 1024;
            ubyte[BUFSIZE] buf;
            int l = data.read(buf);
            messageBody = "";
            messageBody ~= cast(char[])buf[0..l];
            while (l == BUFSIZE) {
                l = data.read(buf);
                messageBody ~= cast(char[])buf[0..l];
            }
        }
        else {
            messageBody = data.readString(len);
        }
    }

    char[] toString() {
        if (all) return all;

        all = statusLine ~ "\r\n";

        char[][] keys = header.keys;

        foreach (char[] k; header.keys) {
            all ~= k ~ ": " ~ header[k] ~ "\r\n";
        }

        all ~= "\r\n" ~ messageBody;

        return all;
    }

public:
    char[] all;
    char[] statusLine;
    char[] httpVersion;
    char[] code;
    char[] message;
    char[][char[]] header;
    char[] messageBody;
}

class Http {
    this(char[] host, int port) {
        host_ = host;
        port_ = port;
    }

    abstract HttpResponse get_(char[] uri);

    HttpResponse get(char[] uri) { return get_(uri); }
    HttpResponse get(char[] uri, char[] q) {
        return get(uri ~ "?" ~ q);
    }
    HttpResponse get(char[] uri, char[][char[]] q) {
        return get(uri, joinQuery(q));
    }

    abstract HttpResponse post_(char[] uri, char[] q);

    HttpResponse post(char[] uri, char[] q) { return post_(uri, q); }
    HttpResponse post(char[] uri, char[][char[]] q) {
        return post(uri, joinQuery(q));
    }

protected:
    char[] joinQuery(char[][char[]] q) {
        char[] qs;
        foreach (char[] k, char[] v; q) {
            qs ~= k ~ "=" ~ v ~ "&";
        }
        qs.length = qs.length - 1;
        return std.uri.encode(qs);
    }

    char[] createCommonHeader() {
        char[] headerStr = "Host: " ~ host_ ~ "\r\n";
        foreach (char[] k, char[] v; header) {
            headerStr ~= k ~ ": " ~ v ~ "\r\n";
        }
        return headerStr;
    }

    SocketStream connect() {
        Socket sock = new Socket(cast(AddressFamily)AddressFamily.INET,
                                 SocketType.STREAM,
                                 ProtocolType.TCP);
        //sock.blocking = false;
        sock.connect(new InternetAddress(host_, port_));
        return new SocketStream(sock);
    }

protected:
    char[] host_;
    int port_;
    char[][char[]] header;
}

class Http09 : Http {
    this(char[] host, int port) {
        super(host, port);
    }

    override HttpResponse get_(char[] uri) {
        char[] req = "GET " ~ uri ~ "\r\n";

        SocketStream ss = connect();

        ss.writeString(req);

        HttpResponse res = new HttpResponse(ss);

        ss.close();

        return res;
    }

    override HttpResponse post_(char[] uri, char[] q) {
        // not implemented
        assert(false);
        return null;
    }
}

class Http10 : Http {
    this(char[] host, int port) {
        super(host, port);
    }

    override HttpResponse get_(char[] uri) {
        char[] req =
            "GET " ~ uri ~ " HTTP/1.0\r\n" ~
            createCommonHeader() ~ "\r\n";

        SocketStream ss = connect();

        ss.writeString(req);

        HttpResponse res = new HttpResponse(ss);

        ss.close();

        return res;
    }

    override HttpResponse post_(char[] uri, char[] q) {
        char[] req =
            "POST " ~ uri ~ " HTTP/1.0\r\n" ~
            createCommonHeader() ~
            "Content-Length: " ~ std.string.toString(q.length) ~ "\r\n"
            "\r\n" ~ q;

        SocketStream ss = connect();

        ss.writeString(req);

        HttpResponse res = new HttpResponse(ss);

        ss.close();

        return res;
    }
}

class Http11 : Http {
    this(char[] host, int port) {
        super(host, port);
        close_ = false;
    }

    override HttpResponse get_(char[] uri) {
        char[] req =
            "GET " ~ uri ~ " HTTP/1.1\r\n" ~
            createCommonHeader() ~
            "Connection: " ~ ((close_) ? "close" : "keep-alive") ~ "\r\n"
            "\r\n";

        if (!ss_) ss_ = connect();

        ss_.writeString(req);

        HttpResponse res = new HttpResponse(ss_);

        if (`Connection` in res.header &&
            std.string.tolower(res.header[`Connection`]) != `keep-alive`) {
            ss_.close();
            ss_ = null;
        }

        return res;
    }

    override HttpResponse post_(char[] uri, char[] q) {
        char[] req =
            "POST " ~ uri ~ " HTTP/1.0\r\n" ~
            createCommonHeader() ~
            "Connection: " ~ ((close_) ? "close" : "keep-alive") ~ "\r\n"
            "Content-Length: " ~ std.string.toString(q.length) ~ "\r\n"
            "\r\n" ~ q;

        if (!ss_) ss_ = connect();

        ss_.writeString(req);

        HttpResponse res = new HttpResponse(ss_);

        if (std.string.tolower(res.header[`Connection`]) != `keep-alive`) {
            ss_.close();
            ss_ = null;
        }

        return res;
    }

    bool alive() { return ss_ !is null; }

    bool close() { return close; }
    void close(bool c) { close_ = c; }

private:
    SocketStream ss_;
    bool close_;
}

unittest {

    Http http;

    http = new Http09("localhost", 80);
    dout.writeLine(http.get("/~i/env.cgi", "hogehoge").toString);

    http = new Http10("localhost", 80);
    dout.writeLine(http.get("/~i/env.cgi", "hogehoge").toString);

    http = new Http10("localhost", 80);
    dout.writeLine(http.post("/~i/env.cgi", "hogehoge").toString);

    Http11 http11;
    http11 = new Http11("localhost", 80);
    dout.writeLine(http11.get("/~i/env.cgi").toString);
//    assert(http11.alive);
    http11.close = true;
    dout.writeLine(http11.get("/~i/env.cgi").toString);
//    assert(!http11.alive);
}

